% Read the individual body images, compute the feature (3D-Histogram) and
% perform agglomerative clustering

clear;
clc;
close all;

%% Episode handling
epinum = '6';

fprintf('Episode %s\n',epinum);
% paths to load images and corresponding maps
imDbPath = ['/home/mtapaswi/Database/bodyImages/E0',epinum,'/'];
imMapPath = ['/home/mtapaswi/Database/bodyImages/Map_E0',epinum,'/'];

% read the cuts
cutLabels = dlmread(['/home/mtapaswi/MS_Thesis/FinalLabels/cuts_e',epinum]);

%% Setting Parameters

verbose = false;

% filenamessaveFeatFile
FeatFilename = ['/home/mtapaswi/MS_Thesis/Generated_Data/Clothing/ClothingFeatures/bodyImFeats_E0',epinum,'.mat'];

% check whether to compute histograms or just load...
if exist(FeatFilename,'file')
    fprintf('*** Feature file exists. Loading... ***\n');
    load(FeatFilename);
    computeHist = false;
else
    fprintf('*** Feature file does NOT exist. Computing... ***\n');
    computeHist = true;
end

% 3D histogram - 4x4x4 bins
bins3D = 4; 
useHSV = false;
truncateHist = 0.1;

% Clothing location (estimated)
leftcut = 0.1; % cut the left 10% of image (bg)
rightcut = 0.9; % cut the right 10% of image (bg)
topcut = 0.2; % cut the top 20% of image (face)
botcut = 0.5; % cut the bottom 50% of image (legs/bg)
useArea = 0.25; % atleast 25% of area should be made up of useful image pixels

%% Load all images and compute the feature

allIms = dir([imDbPath,'*.png']);

if computeHist    
    orderFrames = [1:length(allIms); zeros(1,length(allIms))];
    hist3Dim = zeros(length(allIms), bins3D^3);
    skipImage = zeros(length(allIms), 1);
    
    for k=1:length(allIms)
        filename = allIms(k).name;

        % parse the filename to get information
        underscores = strfind(filename, '_');
        id = filename(1:underscores(1)-1);
        frameNum = str2double(filename(underscores(1)+1:underscores(2)-1));
        orderFrames(2,k) = frameNum;

        % get basic info
        fprintf('%4d. Processing %s\n',k,filename);
        bodyIm(k).fileName = filename;
        bodyIm(k).frameNum = frameNum;
        bodyIm(k).Id = id;
        
        % read image and the map (mask)
        imin = imread([imDbPath, filename]);
        mapin = imread([imMapPath, filename]);

        % cut to estimated clothing location
        useIm1 = imin(round(topcut*end):round(botcut*end),round(leftcut*end):round(rightcut*end),:);
        useMap1 = mapin(round(topcut*end):round(botcut*end),round(leftcut*end):round(rightcut*end),:);
        
        % define skip image conditions
        choppedArea = numel(useMap1); % chopping done for selecting body bounding box
        croppedArea = sum(sum(useMap1)); % cropping done because part of the box was out of frame
        skipImage(k) = (croppedArea/choppedArea) < useArea;

        if skipImage(k)
            fprintf('----> Skipped, area is %.3f\n',croppedArea/choppedArea);
        else
            % select the useful parts (actually the white parts of the mask)
            useIm2 = imageCropper(useIm1, useMap1);
            
            % compute the 3D histogram and store for clustering...
            hist3Dim(k,:) = normalizeHist3D(hist3D(useIm2, bins3D, useHSV),truncateHist);
                        
            % plotting stuff for debug
            if verbose
                subplot(221); imshow(useIm2);
                subplot(222); stem(bodyIm(k).hist3Dim); 
                drawnow;
            end
        end
    end
    
    save(FeatFilename,'hist3Dim','bodyIm','orderFrames','skipImage');
end


%% Order the image frame number and file number

fprintf('Getting frame number to file number links...\n');
[sortedFrames, idxFiles] = sort(orderFrames(2,:));
skippedImages = find(skipImage==1);


%% Episode "Scening"

fprintf('Performing Scening...\n');
% divide the entire episode using the special sequence / fade transitions / title song divisions
sceneChangeLocs = sort([find(cutLabels(:,1)==1); find(cutLabels(:,1)==2); find(cutLabels(:,1)==3)]);
newSceneLocs = sceneChangeLocs(diff(sceneChangeLocs)==1);
newSceneLocs = [newSceneLocs; size(cutLabels,1)]; % start of new scene, and last shot number to take into account last scene...
frameSceneLocs = [1; cutLabels(newSceneLocs, 5)];


%% Clustering

linkCut = 0.06; % percentage of max distance at which to cut off the clustering process

for k=1:length(frameSceneLocs)-1
    fprintf('*** Clustering Scene %d...\n',k);
    fprintf('Scene is from frame %d to frame %d\n',frameSceneLocs(k),frameSceneLocs(k+1));
    
    % choose the histograms to be clustered, 
    pickSortedFrames = intersect(find(sortedFrames>frameSceneLocs(k)),find(sortedFrames<frameSceneLocs(k+1)));
    pickHist = idxFiles(pickSortedFrames);
    
    % remove the skipped ones
    sceneImages{k} = setdiff(pickHist, skippedImages); % the allIms(finalList).name accessible list...
    clusterHist = hist3Dim(sceneImages{k},:);
    
    % seems like some all 0 histogram has been picked! :(
    if any(all(~clusterHist,2)), keyboard; end
    
    % start clustering process...
    fprintf('\tClustering %d images...\n',size(clusterHist,1));
    
    fprintf('Computing distance...');
    distMat = pdist(clusterHist,'euclidean');
    fprintf(' Done\n');

    fprintf('Computing linkages...');
    linkages = linkage(distMat,'ward');
    fprintf(' Done\n');

    % compute distance threshold to stop hierarchical clustering process
    cutOffDistance = linkCut*max(linkages(:,3));
%     cumsumLinkages = cumsum(linkages(:,3));
%     cumsumCutOff = sumLinkCut*max(cumsumLinkages);
%     leftOut = sum(cumsumLinkages>cumsumCutOff);
%     newCutOffDistance = linkages(end-leftOut,3)+eps;
    
    fprintf('Computing clusters with cutoff distance = %.3f...',cutOffDistance);
    imageClass{k} = cluster(linkages,'criterion','distance','cutoff',cutOffDistance); 
    fprintf(' %d Done\n',max(imageClass{k}));
    
    % print numclasses and assignment
    for class=1:max(imageClass{k})
        fprintf('%d <- %d || ',class,sum(imageClass{k}==class));
    end
    fprintf('\n');

    fprintf('Cophenetic correlation coefficient is %.3f\n',cophenet(linkages,distMat));

    % Cluster goodness visualization
%     clf
%     subplot(221), plot(linkages(:,3)); hold on; plot(cutOffDistance*ones(1,size(linkages,1)),'r'); plot(newCutOffDistance*ones(1,size(linkages,1)),'k'); title('Distance with cluster number');
%     subplot(222), plot(cumsumLinkages); hold on; plot(cumsumCutOff*ones(1,size(linkages,1)),'r');title('Cumulative sum of linkage distances');
%     subplot(223), 
% dendrogram(linkages,'colorthreshold',3); %title('Dendrogram with normally computed cutoff');
% set(gca, 'XTick',[]); ylabel('Distance');

%     subplot(224), dendrogram(linkages,'colorthreshold',newCutOffDistance); title('Dendrogram with Cumulative computed cutoff');
%     drawnow;
%     pause
% keyboard;
    fprintf('----------------------------------------------------------------------------------\n');
end
    

